------------------------------------------------
--                CT_BuffMod                  --
--                                            --
-- Mod that allows you to heavily customize   --
-- the display of buffs to your liking.       --
-- Please do not modify or otherwise          --
-- redistribute this without the consent of   --
-- the CTMod Team. Thank you.                 --
------------------------------------------------

local _G = getfenv(0);
local module = _G.CT_BuffMod;

local BUFF = "HELPFUL|PASSIVE";
local DEBUFF = "HARMFUL";
local TOOLTIP = CreateFrame("GameTooltip", "CT_BuffModTooltip", nil, "GameTooltipTemplate");
local TOOLTIP_TITLE = _G.CT_BuffModTooltipTextLeft1;
TOOLTIP:SetOwner(WorldFrame, "ANCHOR_NONE");

local DELTA_TIMELEFT_THRESHOLD = 0.5;
local AURA_UPDATE_FREQUENCY = 1.0;

--------------------------------------------
-- Local copies

local GetPlayerBuff = GetPlayerBuff;
local GetPlayerBuffName = GetPlayerBuffName;
local GetPlayerBuffApplications = GetPlayerBuffApplications;
local GetPlayerBuffTexture = GetPlayerBuffTexture;
local GetPlayerBuffTimeLeft = GetPlayerBuffTimeLeft;
local GetWeaponEnchantInfo = GetWeaponEnchantInfo;
local strmatch = strmatch;

--------------------------------------------
-- Gather data

local auraTable = { };

local mainHandEnchant, offHandEnchant;
local showItemBuffs = true;

local auratypeBUFF = 1;
local auratypeDEBUFF = 2;
local auratypeITEM = 3;

module.getAuraTable = function()
	return auraTable;
end

local function getAuraApplications(auraIndex)
	local tblEntry = auraTable[auraIndex];
	local _, count;
	if (tblEntry.bdType == auratypeBUFF) then
		_, _, _, count = UnitBuff(module.auraUnit, tblEntry.bdIndex);
	elseif (tblEntry.bdType == auratypeDEBUFF) then
		_, _, _, count = UnitDebuff(module.auraUnit, tblEntry.bdIndex);
	else
		count = 0;
	end
	if (not count) then
		count = 0;
	end
	return count;
end

local function getAuraDispelType(auraIndex)
	local tblEntry = auraTable[auraIndex];
	local _, debuffType;
	if (tblEntry.bdType == auratypeBUFF) then
		_, _, _, _, debuffType = UnitBuff(module.auraUnit, tblEntry.bdIndex);
	elseif (tblEntry.bdType == auratypeDEBUFF) then
		_, _, _, _, debuffType = UnitDebuff(module.auraUnit, tblEntry.bdIndex);
	else
		debuffType = "";
	end
	if (not debuffType) then
		debuffType = "";
	end
	return debuffType;
end

module.getAuraApplications = getAuraApplications;
module.getAuraDispelType = getAuraDispelType;

local function getAuraTimeLeft(bdType, bdIndex)
	local _, duration, expirationTime, timeLeft;
	if (bdType == auratypeBUFF) then
		_, _, _, _, _, duration, expirationTime = UnitBuff(module.auraUnit, bdIndex);
	elseif (bdType == auratypeDEBUFF) then
		_, _, _, _, _, duration, expirationTime = UnitDebuff(module.auraUnit, bdIndex);
	else
		expirationTime = GetTime();
		duration = expirationTime;
	end
	if (not expirationTime) then
		expirationTime = GetTime();
		duration = expirationTime;
	end
	timeLeft = expirationTime - GetTime();
	if (timeLeft < 0) then
		timeLeft = 0;
	end
	return timeLeft, duration;
end

local function removeAura(auraIndex, tblEntry)
	-- ----------
	-- Remove an aura from auraTable.
	--
	-- Params:
	--   auraIndex - Index into auraTable of the item to be removed.
	-- ----------
	tremove(auraTable, auraIndex);
	tblEntry:drop();
end

--[[ local function getBuffType(auraIndex, newBuffId, timeLeft)
	timeLeft = timeLeft or GetPlayerBuffTimeLeft(auraIndex);
	if ( newBuffId ) then
		if ( GetPlayerBuff(newBuffId, BUFF) == auraIndex ) then
			if ( timeLeft == 0 ) then
				return "AURA";
			else
				return "BUFF";
		else
			return "DEBUFF";
		end
	else
		-- Loop through all buffs
	end
end ]]

local function addAura(auraIndex, bdType, texture, timeLeft, name, bdIndex, duration)
	-- ----------
	-- Add an aura to auraTable.
	--
	-- Params:
	--   auraIndex - Index into auraTable of the item to be added.
	--   bdType    - Type of aura being added (auratypeBUFF or auratypeDEBUFF).
	--   texture   - Icon texture file name.
	--   timeLeft  - Time left on the aura in seconds.
	--   name      - Name of the aura.
	--   bdIndex   - Index value used in the UnitBuff() or UnitDebuff() function.
	--   duration  - Total duration of the aura in seconds.
	-- ----------
	local tblEntry = module:getBuffObject();

	tblEntry.name = name;
	tblEntry.texture = texture;
	tblEntry.timeleft = timeLeft;
	tblEntry.bdIndex = bdIndex;
	tblEntry.bdType = bdType;
	tblEntry.duration = duration;
	
	if (bdType == auratypeBUFF) then
		if ( timeLeft == 0 ) then
			tblEntry.type = "AURA";
		else
			tblEntry.type = "BUFF";
		end
	else
		tblEntry.type = "DEBUFF";
	end
	
	tinsert(auraTable, tblEntry);
	tblEntry:add(auraIndex, bdType, bdIndex, nil, duration);
end

local function isAuraInThreshold(timeLeft, tblEntry)
	return abs(timeLeft - tblEntry.timeleft) <= AURA_UPDATE_FREQUENCY;
end

local function scanAuras2(auraIndex, bdType)
	-- ----------
	-- Scan the auras on the player and update the auraTable.
	--
	-- Params:
	--   auraIndex - Next index to update in auraTable.
	--   bdType    - Type of aura to scan for (auratypeBUFF or auratypeDEBUFF).
	--
	-- Returns:
	--   1) Next index to udpate in auraTable.
	--   2) Number of auras + 1 that were found on the player.
	-- ----------
	local isIdentical, tblEntry, timeLeft, readNext;
	local bdIndex = 1;

	-- Determine which function we'll use to get aura information.
	local func;
	if (bdType == auratypeBUFF) then
		func = UnitBuff;
	else
		func = UnitDebuff;
	end

	-- Get information about the first aura.
	local name, rank, texture, count, debuffType, duration, expirationTime = func(module.auraUnit, bdIndex);

	-- Process all auras on the player.
	while (name) do
		readNext = true;

		-- Calculate the time remaining on this aura.
		timeLeft = expirationTime - GetTime();
		if (timeLeft < 0) then
			timeLeft = 0;
		end

		-- Do we already have an aura saved at this spot in auraTable?
		tblEntry = auraTable[auraIndex];
		if (tblEntry) then
			-- Is this the same aura we saw during the last scan?
			isIdentical = ( tblEntry.texture == texture and tblEntry.name == name );
			if (isIdentical) then
				-- Same aura as before.
				tblEntry.bdIndex = bdIndex;
				auraIndex = auraIndex + 1;
			else
				-- Aura is different than the last scan. Remove it from the list.
				removeAura(auraIndex, tblEntry);
				if ( auraIndex > #auraTable ) then
					-- We just removed the last aura in the list.
					-- Add the new aura to the end of the table.
					addAura(auraIndex, bdType, texture, timeLeft, name, bdIndex, duration);
					auraIndex = auraIndex + 1;
				else
					-- Will need to re-scan the current aura because we removed
					-- the item at this spot in auraTable causing the rest to shift.
					readNext = false;
				end
			end
		else
			-- Add the aura to the end of the table.
			addAura(auraIndex, bdType, texture, timeLeft, name, bdIndex, duration);
			auraIndex = auraIndex + 1;
		end

		if (readNext) then
			-- Get information for the next aura.
			bdIndex = bdIndex + 1;
			name, rank, texture, count, debuffType, duration, expirationTime = func(module.auraUnit, bdIndex);
		end
	end
	-- Return the next spot in auraTable to update if there are any more buffs/debuffs to examine.
	return auraIndex;
end

local function scanAuras()
	-- ----------
	-- Scan for buffs and debuffs on the player.
	-- ----------
	local tblEntry;
	local auraIndex = 1;

	-- Scan for buffs
	auraIndex = scanAuras2(auraIndex, auratypeBUFF);

	-- Scan for debuffs
	auraIndex = scanAuras2(auraIndex, auratypeDEBUFF);

	-- If there are fewer auras now than there were before, then delete the extra in auraTable.
	while (auraIndex <= #auraTable) do
		tblEntry = auraTable[auraIndex];
		removeAura(auraIndex, tblEntry);
	end
end

local function getWeaponEnchantInfo(isMainHand)
	local slot = ( isMainHand and 16 ) or 17;
	
	TOOLTIP:ClearLines();
	TOOLTIP:SetInventoryItem(module.auraUnit, slot);
	local text, name;
	for i=1, TOOLTIP:NumLines(), 1 do
		text = _G["CT_BuffModTooltipTextLeft"..i]:GetText();
		name = strmatch(text, "^.+ %(%d+%s+.+%)$");
		if ( name ) then
			break;
		end
	end
	
	return name or "Weapon Buff", GetInventoryItemTexture(module.auraUnit, slot), slot;
end

local function addWeaponEnchant(isMainHand, expiration, charges)
	local tblEntry = module:getBuffObject();
	
	local slot;
	tblEntry.name, tblEntry.texture, slot = getWeaponEnchantInfo(isMainHand);
	tblEntry.timeleft = expiration/1000;
	tblEntry.type = "ITEM";
	tblEntry.itemEnchantId = isMainHand and 1 or 2;
	tblEntry.duration = tblEntry.timeleft;
	tblEntry:add(slot, auratypeITEM, 0, charges, tblEntry.timeleft);
	
	return tblEntry;
end

local function updateWeaponEnchantments(difference)
	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo();
	local oldTime;
	local isPlayer = (module.auraUnit == "player");
	
	-- Main Hand
	if ( hasMainHandEnchant and showItemBuffs and isPlayer ) then

		if ( not mainHandEnchant ) then
			mainHandEnchant = addWeaponEnchant(true, mainHandExpiration, mainHandCharges);
		end

		mainHandExpiration = ( mainHandExpiration or 0 )/1000;
		oldTime = mainHandEnchant.timeleft;
		mainHandEnchant.timeleft = mainHandExpiration;
		if ( abs(mainHandExpiration-oldTime) > difference ) then
			-- Weapon Enchantment was renewed
			mainHandEnchant.name, mainHandEnchant.texture = getWeaponEnchantInfo(true);
			mainHandEnchant:renew(mainHandExpiration, mainHandCharges, mainHandExpiration);
		else
			mainHandEnchant:updateTimeDisplay(mainHandExpiration, mainHandCharges);
		end
	elseif ( mainHandEnchant ) then
		mainHandEnchant:drop();
		mainHandEnchant = nil;
	end

	-- Off Hand
	if ( hasOffHandEnchant and showItemBuffs and isPlayer ) then

		if ( not offHandEnchant ) then
			offHandEnchant = addWeaponEnchant(false, offHandExpiration, offHandCharges);
		end

		offHandExpiration = ( offHandExpiration or 0 )/1000;
		oldTime = offHandEnchant.timeleft;
		offHandEnchant.timeleft = offHandExpiration;
		if ( abs(offHandExpiration-oldTime) > difference ) then
			-- Weapon Enchantment was renewed
			offHandEnchant.name, offHandEnchant.texture = getWeaponEnchantInfo(false);
			offHandEnchant:renew(offHandExpiration, offHandCharges, offHandExpiration);
		else
			offHandEnchant:updateTimeDisplay(offHandExpiration, offHandCharges);
		end
	elseif ( offHandEnchant ) then
		offHandEnchant:drop();
		offHandEnchant = nil;
	end
end

local function getUnit()
	local unit;
	if (UnitHasVehicleUI("player")) then
		unit = "vehicle";
	else
		unit = "player";
	end
	return unit;
end

local function updateAuraDurations(difference)
	difference = AURA_UPDATE_FREQUENCY - difference + DELTA_TIMELEFT_THRESHOLD;
	local newTime, duration;
	
	if (not module.auraUnit) then
		return;
	end

	-- Auras
	local oldTime;
	for key, value in ipairs(auraTable) do
		newTime, duration = getAuraTimeLeft(value.bdType, value.bdIndex);
		oldTime = value.timeleft;
		value.timeleft = newTime;
		
		if ( newTime > oldTime+difference ) then
			-- Buff was renewed
			value:renew(newTime, nil, duration);
		else
			value:updateTimeDisplay(newTime);
		end
	end
	
	-- Weapon Enchantments
	updateWeaponEnchantments(difference);

end

local function checkAuras(event, arg1, ...)
	local unit = getUnit();
	if ( event == "UNIT_AURA" ) then
		if ( arg1 ~= unit ) then
			return;
		end
	elseif ( event == "UNIT_ENTERED_VEHICLE" or event == "UNIT_EXITED_VEHICLE" ) then
		if ( arg1 ~= "player" ) then
			return;
		end
	end
	module.auraUnit = unit;
	local currentTableIndex = #auraTable;
	scanAuras(1, currentTableIndex + 1);
end

local function checkAuras2(...)
	checkAuras(...);
	updateAuraDurations(0);
end

function module:showItemBuffs(show)
	showItemBuffs = show;
end

module:regEvent("UNIT_AURA", checkAuras);
module:regEvent("PLAYER_ENTERING_WORLD", checkAuras2);
module:schedule(AURA_UPDATE_FREQUENCY, true, updateAuraDurations);
module:regEvent("UNIT_ENTERED_VEHICLE", checkAuras2);
module:regEvent("UNIT_EXITED_VEHICLE", checkAuras2);
